import logging
from shakersynth.synth.rotor import RotorSynth
from shakersynth.config import loader
from typing import Dict  # noqa: F401

config = loader.load_config()
log = logging.getLogger(__name__)
log.setLevel(config.log.level)

helicopters = ['mi-8mt', 'mi-24p', 'uh-1h']
fixed_wings = ['a10-c_2']
all_modules = helicopters + fixed_wings


class Aircraft:
    """An `Aircraft` is the current aircraft/module the player is controlling.

    The `Aircraft` contains instances of `shakersynth.synth` in a list.
    Calling `update` on the `Aircraft` causes each synth to also be updated.
    """
    def __init__(self, module: str):
        """Create a new `Aircraft`, with a collection of synths."""
        self.module = module
        self.synths = []
        self.is_running = False

        if module in helicopters:
            if config.modules[module].effects.rotor.enabled:
                self.synths.append(RotorSynth())
            else:
                log.info(f"Rotor effect disabled for {module}.")

    def update(self, telemetry: dict) -> None:
        """Update all synths with the `telemetry` payload."""
        for synth in self.synths:
            try:
                synth.update(telemetry)
            except NotImplementedError:
                log.error(f"{type(synth)}: unsupported module '{self.module}'")

    def start(self) -> None:
        """Start all synths, enabling audio output."""
        for synth in self.synths:
            log.debug("Starting synth: %s" % type(synth))
            synth.start()
        self.is_running = True

    def stop(self) -> None:
        """Stop all synths, disabling audio output."""
        for synth in self.synths:
            log.debug("Stopping synth: %s" % type(synth))
            synth.stop()
        self.is_running = False



#!/usr/bin/env python3

import logging
import os
import platform
import pyo
import sys
import time
from importlib.resources import read_text
from shakersynth.config.loader import create_default_config_file, load_config
from shakersynth.aircraft.aircraft import Aircraft
from shakersynth.receiver.shakersynth import ShakersynthReceiver

config = load_config()
logging.basicConfig()
log = logging.getLogger("main")
log.setLevel(config.log.level)


def install_lua():
    if not platform.system() == 'Windows':
        log.info('Skipping export script setup: not running on Windows.')
        return

    user_profile = os.path.expandvars('%UserProfile%')
    for dcs in ['DCS', 'DCS.openbeta']:
        script_dir = os.path.join(user_profile, 'Saved Games', dcs, 'Scripts')
        script_file = os.path.join(script_dir, 'Shakersynth.lua')
        if os.path.isdir(script_dir):
            with open(script_file, 'w') as lua:
                try:
                    lua.write(read_text(__package__, 'Shakersynth.lua'))
                except Exception:
                    log.error(f"Failed to write export script: {script_file}")


def main():
    print("\n" + "-" * 35 + "Shakersynth" + "-" * 35)
    create_default_config_file()
    install_lua()

    # Audio synthesis setup.
    server = pyo.Server(
        nchnls=2,
        duplex=0,
        buffersize=config.audio.buffer_size,
        sr=config.audio.sample_rate
    )
    server.setVerbosity(1)

    # Map audio outputs from internal numbers (which can be sparse and large)
    # to a friendly, 1-indexed, monotonic series for the user.
    print("Scanning audio devices...")
    output_devices = pyo.pa_get_devices_infos()[1]
    default_api = pyo.pa_get_default_host_api()
    friendly_to_internal = {}

    print("Found these audio outputs:")
    device_details = enumerate(output_devices.items())
    for friendly_index, (internal_index, properties) in device_details:
        # Only list devices that are available using the default audio API.
        if properties["host api index"] == default_api:
            friendly_index += 1
            friendly_to_internal[friendly_index] = internal_index
            print(f"  {friendly_index}: {properties['name']}")

    print("Enter device ID to use: ", end="")
    chosen_device = int(input())
    server.setOutputDevice(friendly_to_internal[chosen_device])

    server.boot()
    server.start()
    server.setAmp(config.audio.global_volume)

    # Receive Telemetry from the Shakersynth DCS export script.
    receiver = ShakersynthReceiver()

    print("*" * 80)
    print("ShakerSynth now running.")
    print("Press ctrl+c to exit.")
    print("*" * 80)
    log.info("Waiting for initial telemetry.")

    # A little state machine to keep track of our aircraft and telemetry.
    running = False
    while True:
        telemetry = receiver.receive()

        if telemetry and not running:
            # Then we have just started a mission.
            module = telemetry['module']
            log.info(f"Starting new aircraft: {module}")
            aircraft = Aircraft(module)
            aircraft.start()
            aircraft.update(telemetry)
            running = aircraft.is_running

        elif telemetry and running:
            # Then things are ticking along nicely.
            aircraft.update(telemetry)

        elif running and not telemetry:
            # Then we got an empty telemetry object while running an aircraft.
            # This signifies that we left the mission or lost contact with DCS.
            log.info(f"Shutting down aircraft: {aircraft.module}")
            aircraft.stop()
            running = aircraft.is_running
            time.sleep(0.1)
            del(aircraft)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(0)



import logging
import os
import platform
from typing import Callable

from config import ConfigurationSet, config_from_env, config_from_yaml
from textwrap import dedent


default_yaml = dedent(
    """
    ---
    # Audio and sound card settings.
    audio:
      # Be sure to use the sample rate that your sound card expects.
      sample_rate: 44100

      # If you get audio glitches, try increasing the buffer to 4096.
      buffer_size: 2048

      # Set the overall volume between 0 and 1.0.
      #
      # Values above 0.7 can cause clipping distortion (at least on
      # on the author's system).
      global_volume: 0.7

    # Logging options.
    log:
      # Set level to "debug" for very verbose logging.
      level: info

    # Configuration of individual modules and effects.
    modules:
      uh-1h:
        effects:
          rotor:
            enabled: true
      mi-24p:
        effects:
          rotor:
            enabled: true
      mi-8mt:
        effects:
          rotor:
            enabled: true
    """
    ).strip()

if platform.system() == 'Windows':
    config_dir = os.path.join(
        os.environ['LOCALAPPDATA'],
        'Shakersynth'
    )
else:
    config_dir = os.path.join(
        os.environ['HOME'],
        '.shakersynth'
    )


def get_config_file_path() -> str:
    default = os.path.join(config_dir, 'shakersynth.yml')
    return os.getenv("SHAKERSYNTH_CONFIG_FILE", default)


def create_default_config_file() -> None:
    config_file_path = get_config_file_path()
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)

    if not os.path.exists(config_file_path):
        with open(config_file_path, 'w') as config_file:
            config_file.write(default_yaml)


def mutate(conf: dict, key: str, func: Callable):
    """Change the value of `conf[key]` by passing it to `func`."""
    conf[key] = func(conf[key])
    return conf


def load_config():
    create_default_config_file()
    with open(get_config_file_path()) as f:
        cfg = ConfigurationSet(
            config_from_env("SHAKERSYNTH", separator="_", lowercase_keys=True),
            config_from_yaml(f.read()),
            config_from_yaml(default_yaml))

    cfg = mutate(cfg, "audio.global_volume", lambda x: float(x))
    cfg = mutate(cfg, "log.level", lambda x: getattr(logging, x.upper()))

    return cfg



import logging
import pyo
from shakersynth.config import loader

config = loader.load_config()
log = logging.getLogger(__name__)
log.setLevel(config.log.level)


class RotorSynth():
    """Effect synthesizer for helicopter rotor vibrations.

    This a stereo (2 channel) effect.
    """
    def __init__(self):
        """Create a new RotorSynth."""
        self.is_running = False

        # Create two volume faders which we'll use to control
        # the volume of the effect.
        self.fader0 = pyo.Fader()
        self.fader1 = pyo.Fader()

        # These Lorenz attractors create randomness, which we will use to
        # modulate the signal a little. It adds character and makes the
        # effect feel less repetitive.
        self.lorenz0 = pyo.Lorenz(pitch=0.1)
        self.lorenz0_scaled = pyo.Scale(self.lorenz0, outmin=0.3, outmax=0.9)
        self.lorenz1 = pyo.Lorenz(pitch=0.11)  # Slightly different for stereo.
        self.lorenz1_scaled = pyo.Scale(self.lorenz1, outmin=0.3, outmax=0.9)

        # These are the actual signal sources. Volume is controlled by the
        # fader objects and the "sharpness" of the wave is modulated by the
        # Lorenz attractors.
        self.osc0 = pyo.LFO(
            freq=0.0001,
            sharp=self.lorenz0_scaled,
            mul=self.fader0)
        self.osc1 = pyo.LFO(
            freq=0.0001,
            sharp=self.lorenz1_scaled,
            mul=self.fader1)

        # Then we'll filter the signals so that only low frequencies are
        # sent to the shakers.
        self.filter0 = pyo.Biquad(self.osc0, freq=200)
        self.filter1 = pyo.Biquad(self.osc1, freq=200)

        # Finally, hook the output of the filters up to the sound card's left
        # and right channels.
        self.filter0.out(0)
        self.filter1.out(1)

    def update(self, telemetry: dict) -> None:
        """Update synth parameters based on `telemetry`.

        Each time `update` is called with a `telemetry` payload, the
        `RotorSynth` will adjust its synthesis parameters to match.

        `telemetry` must contain the key `rotor_rpm_percent` to have
        any effect.

        Supports the Mi-8, Mi-24, and UH-1H modules.
        """
        module = telemetry["module"]
        if(module == "mi-8mt" or module == "mi-24p"):
            blade_count = 5
        elif(module == "uh-1h"):
            blade_count = 2
        else:
            raise NotImplementedError

        rpm = self._calculate_rotor_rpm(telemetry)

        # Protect against divide-by-zero errors.
        if rpm == 0:
            rpm = 0.00000001

        # Revolutions per second is more useful than RPM.
        revolutions_per_second = rpm / 60.0

        blades_per_second = revolutions_per_second * blade_count
        log.debug(f"blades_per_second: {blades_per_second}")
        self.osc0.setFreq(blades_per_second)
        self.osc1.setFreq(blades_per_second)

        # Only apply the Lorenz attractor "pixie dust" when the
        # rotor is actually moving.
        log.debug(f"Lorenz min: {min(0.3, blades_per_second)}")
        log.debug(f"Lorenz max: {min(0.9, blades_per_second)}")
        self.lorenz0_scaled.setOutMin(min(0.3, blades_per_second))
        self.lorenz0_scaled.setOutMax(min(0.9, blades_per_second))
        self.lorenz1_scaled.setOutMin(min(0.3, blades_per_second))
        self.lorenz1_scaled.setOutMax(min(0.9, blades_per_second))

    def start(self) -> None:
        """Start the `RotorSynth`, activating audio output."""
        if self.is_running:
            return
        log.debug("Started.")
        self.fader0.play()
        self.fader1.play()
        self.is_running = True

    def stop(self) -> None:
        """Stop the `RotorSynth`, deactivating audio output."""
        if not self.is_running:
            return
        log.debug("Stopped.")
        self.fader0.stop()
        self.fader1.stop()
        self.is_running = False

    def _calculate_rotor_rpm(self, telemetry: dict) -> float:
        """Given a telemetry payload, return the true RPM of the rotor."""
        module = telemetry["module"]
        rpm_percent = telemetry["rotor_rpm_percent"]

        if(module == "mi-8mt"):
            # 95 gauge RPM == 192 real rotor RPM. [1, 2]
            # But 200 gives better results in the sim.
            return float(rpm_percent * (200 / 95))
        elif(module == "mi-24p"):
            # 280 gives good sync on the Hind, but that seems way off??
            return float(rpm_percent * (280 / 95))
        elif(module == "uh-1h"):
            # 90 gauge RPM == 324 real rotor RPM. [3]
            return float(rpm_percent * (324 / 90))
        else:
            return 0.00000001

# 1. http://koavia.com/eng/product/helicopter/hvostovye_valy.shtml#2
# 2. https://www.pprune.org/rotorheads/221789-mil-8-mtv-mtv-1-info.html
# 3. https://apps.dtic.mil/dtic/tr/fulltext/u2/901787.pdf
